#!/usr/bin/python

import argparse
import json
from http.server import HTTPServer, BaseHTTPRequestHandler
import hashlib
import urllib.parse
from datetime import datetime
import random


global asset_idx


class Asset(object):

    NOT_COMPLETED = 'not_completed'
    COMPLETED = 'completed'
    SKIPPED = 'skipped'

    def __init__(self,
                 idx,
                 status=NOT_COMPLETED):
        self.status = status
        self.req_count = 0
        self.token = None
        self.reason = None
        self.idx = idx


tokens_map = {}


def append_image_info(token, js={}):
    status = tokens_map[token].status
    idx = tokens_map[token].idx

    res = {
        'token': token,
        'status': status,
        'idx': idx
    }

    for k in js:
        if js[k] is None:
            continue
        res[k] = js[k]

    if status == Asset.COMPLETED:
        res['name'] = 'dummy subject %d' % (idx)

        # 80% chance that the image has structured_output
        if random.randint(0, 10) < 8:
            res['structured_output'] = {
                'quantity': [
                    'quantity_%d' % i for i in range(2) if i < random.randint(0, 2)],
                'gender': [
                    'gender_%d' % i for i in range(2) if i < random.randint(0, 2)],
                'material': [
                    'material_%d' % i for i in range(4) if i < random.randint(0, 4)],
                'color': [
                    'color_%d' % i for i in range(3) if i < random.randint(0, 3)]
            }

        # 80% chance that the image has categories
        if random.randint(0, 10) < 8:
            res['categories'] = [
                'category_%d' % i for i in range(4) if i < random.randint(0, 4)]

        return res

    if status == Asset.SKIPPED:
        res['reason'] = 'dummy reason %d' % (idx)
        return res

    return res


class S(BaseHTTPRequestHandler):

    def _set_headers(self, status_code=200):
        self.send_response(status_code)
        self.send_header('Content-type', 'application/json')
        self.end_headers()

    def response(self, json_message, status_code=200):
        self._set_headers(status_code=status_code)
        print(json.dumps(json_message, indent=2))
        self.wfile.write(json.dumps(json_message).encode('utf8'))

    def do_GET(self):

        p = self.path
        while p.startswith('/'):
            p = p[1:]
        while p.endswith('/'):
            p = p[:-1]
        p = p.split('/')
        if len(p) != 3:
            self.response({
                'invalid_path': self.path
            }, 400)
            return

        token = p[2]
        if token not in tokens_map:
            self.response({
                'invalid_token': token
            }, 400)
            return

        tokens_map[token].req_count += 1
        # change response after (random) number of requests
        if tokens_map[token].req_count >= random.randint(0, 6):
            # 10% chance that the image is skipped
            if random.randint(0, 10) < 1:
                tokens_map[token].status = Asset.SKIPPED
            else:
                tokens_map[token].status = Asset.COMPLETED

        # 20% chance that the request does not return 200
        status_code = 200
        if random.randint(0, 5) < 1:
            status_code = 400

        self.response(append_image_info(token), status_code)

    def do_HEAD(self):
        self._set_headers()

    def do_POST(self):

        global asset_idx

        content_len = int(self.headers.get('Content-Length'))
        post_body = self.rfile.read(content_len)

        token = hashlib.sha1(str(datetime.now()).encode('utf-8') +
                             str(post_body[:100]).encode('utf-8')).hexdigest()

        tokens_map[token] = Asset(asset_idx)
        asset_idx += 1

        # 40% chance that the image is immediately analysed
        if random.randint(0, 10) < 4:
            tokens_map[token].status = Asset.COMPLETED

        # 10% chance that the image is immediately skipped
        if random.randint(0, 10) < 1:
            tokens_map[token].status = Asset.SKIPPED

        # 20% chance that the request does not return 200
        status_code = 200
        if random.randint(0, 5) < 1:
            status_code = 400

        self.response(append_image_info(token), status_code=status_code)


def run(server_class=HTTPServer, handler_class=S, addr='localhost', port=8099):
    httpd = server_class((addr, port), handler_class)
    print(f'Starting httpd server on {addr}:{port}')
    httpd.serve_forever()


parser = argparse.ArgumentParser(description='Run a simple HTTP server')
parser.add_argument(
    '-l',
    '--listen',
    default='localhost',
    help='Specify the IP address on which the server listens',
)
parser.add_argument(
    '-p',
    '--port',
    type=int,
    default=8099,
    help='Specify the port on which the server listens',
)
args = parser.parse_args()

if __name__ == '__main__':

    global asset_idx
    asset_idx = 1

    try:
        run(addr=args.listen, port=args.port)
    except KeyboardInterrupt:
        print('stopped by user')
