# coding=utf8

import json
import time
from datetime import datetime as dt
import requests
from threading import Thread

from context import get_json_value
from util import dumpjs, print_traceback, format_datetime, is_string_set, AssetSkippedException
from search import build_search_query, do_search, map_search_results, build_linked_object_search_query
from eas import map_assets, get_asset_info
from easydb_api import easydb_post_objects
from event import log_object_events, log_event, AUTO_KEYWORDER_OBJECT_UPDATE, AUTO_KEYWORDER_OBJECT_CREATE, AUTO_KEYWORDER_CLOUDSIGHT_REQUEST, AUTO_KEYWORDER_ERROR
from cloudsight_config import cloudsight_api_languages


class CloudsightWorker(object):

    def __init__(self,
                 easydb_context,
                 easydb_server,
                 easydb_token,
                 config,
                 search_chunk_size):

        self.easydb_context = easydb_context
        self.easydb_server = easydb_server
        self.easydb_token = easydb_token
        self.logger = easydb_context.get_logger(
            'pf.plugin.base.auto_keyworder.cloudsight.worker')
        self.api_logger = easydb_context.get_logger(
            'pf.plugin.base.auto_keyworder.easydb_api')
        self.config = config
        self.search_chunk_size = search_chunk_size

        # map system_object_id --> util.AssetStatus
        self.asset_status_map = {}

    def format_asset_status(self):
        d = {}
        for k in self.asset_status_map:
            d[k] = self.asset_status_map[k].to_dict()
        return dumpjs(d)

    def main_loop(self):
        # iterate over all different configurations
        # perform searches for affected objects, call cloudsight.ai and update objects
        for c in self.config.cloudsight_config_list:
            if not c.enabled:
                self.logger.debug(
                    'cloudsight config %s not enabled, skip' % c.debug_name)
                continue

            # search all objects of the objecttype where the tagfilter matches and
            # where there is no last keyword timestamp, or where the keyword timestamp is older than the max age

            linked_object_ids = {}

            offset = 0
            has_more = True
            while has_more:
                self.logger.debug('search batch: offset %d, size %d' %
                                  (offset, self.search_chunk_size))
                query = build_search_query(
                    c.objecttype,
                    c.tagfilter,
                    c.asset_field.name,
                    c.timestamp_field.name,
                    min_age_days=c.min_age_days,
                    offset=offset,
                    limit=self.search_chunk_size
                )
                result = do_search(self.easydb_context,
                                   self.easydb_server,
                                   self.easydb_token,
                                   query)

                size_search_results = map_search_results(result,
                                                         c.objecttype,
                                                         self.asset_status_map)
                self.logger.debug(
                    '[%s] result: %d unseen objects' % (c.debug_name, size_search_results))
                if size_search_results < 1:
                    self.logger.debug('[%s] result empty, break search loop')
                    has_more = False
                    break

                # for each object, get the asset data (base64) from the eas and map it to the eas id
                pending_objects = map_assets(self.easydb_context,
                                             self.asset_status_map,
                                             c.objecttype,
                                             c.asset_field.name,
                                             c.asset_version)
                assets_done = set()

                # for each asset data, upload the base64 encoded image to cloudsight and retrieve the keywords
                # and map the keywords to the eas id
                tries = 0
                while tries < self.config.request_status_repititions:
                    self.logger.debug('get keywords for objects (try %d/%d): [%s]' % (tries + 1, self.config.request_status_repititions, ','.join(map(lambda i: str(i), pending_objects))))

                    all_assets_finished = True
                    for sys_id in pending_objects:
                        if sys_id not in self.asset_status_map:
                            continue

                        asset = self.asset_status_map[sys_id]
                        if asset.finished:
                            assets_done.add(sys_id)
                            continue
                        if asset.failed:
                            self.logger.debug('skip requesting of status for object #%d, already failed (reason: %s)' % (
                                sys_id, asset.reason
                            ))
                            continue

                        all_assets_finished = False

                        collector = None
                        if asset.collector is None:
                            self.asset_status_map[sys_id].collector = CloudsightKeywordCollector(
                                self.logger,
                                self.easydb_context,
                                c.api_key,
                                cloudsight_api_languages[c.language],
                                asset.eas_id,
                                c.objecttype,
                                sys_id,
                                asset.base64_data,
                                asset.img_type,
                                asset.original_filepath,
                                asset.uploaded,
                                c.keyword_field_map.keys())

                        subject, keywords, finished, failed, reason = asset.collector.request_keywords(
                            get_json_value(
                                self.asset_status_map[sys_id].easydb_obj, '%s._id' % c.objecttype),
                            c.asset_version
                        )

                        self.asset_status_map[sys_id].uploaded = True
                        self.asset_status_map[sys_id].finished = finished

                        self.asset_status_map[sys_id].failed = failed
                        self.asset_status_map[sys_id].reason = reason

                        self.asset_status_map[sys_id].subject = subject
                        self.asset_status_map[sys_id].keywords = keywords

                    if len(assets_done) > 0:
                        self.logger.debug('successfully got keywords for objects (try %d/%d): [%s]' % (tries + 1, self.config.request_status_repititions, ','.join(map(lambda i: str(i), assets_done))))
                        for sys_id in assets_done:
                            if sys_id in pending_objects:
                                pending_objects.remove(sys_id)

                    if all_assets_finished:
                        break

                    time.sleep(self.config.request_status_delay)
                    tries += 1

                # if there are still any pending assets left, write error events
                for sys_id in pending_objects:
                    event_info = {
                        'error': 'asset was not analyzed in time',
                        'details': {
                            'repititions': self.config.request_status_repititions,
                            'delay': self.config.request_status_delay
                        },
                        'asset': {
                            'eas_id': self.asset_status_map[sys_id].eas_id,
                            'extension': self.asset_status_map[sys_id].img_type,
                            'original_filepath': self.asset_status_map[sys_id].original_filepath,
                            'version': c.asset_version
                        },
                        'object': {
                            '_objecttype': c.objecttype,
                            '_id': get_json_value(
                                self.asset_status_map[sys_id].easydb_obj, '%s._id' % c.objecttype),
                            '_system_object_id': sys_id
                        }
                    }
                    log_event(self.easydb_context,
                              AUTO_KEYWORDER_ERROR,
                              event_info)

                # for all keywords, search the objects with the eas ids and update the objects
                objects_to_update = []
                update_event_infos = []

                # collect all keywords if the target field is a linked field
                # search for already existing linked objects later

                for sys_id in self.asset_status_map:

                    asset = self.asset_status_map[sys_id]
                    if asset.ai_data_added:
                        continue
                    if asset.updated:
                        continue
                    if asset.failed:
                        continue
                    if asset.subject is None and asset.keywords is None:
                        continue

                    subject = asset.subject
                    if subject is not None:
                        if len(subject) < 1:
                            subject = None

                    self.logger.debug('prepare update for object #%d, eas id: %d | subject: %s, keywords: %s' % (
                        sys_id, asset.eas_id, subject, json.dumps(asset.keywords, indent=4)))

                    if asset.keywords is None or len(asset.keywords) < 1:
                        if subject is None or c.subject_field is None:
                            # no data to update. skip
                            continue

                    obj = asset.easydb_obj
                    obj['__keywords'] = {}

                    keywords = []
                    if asset.keywords is not None:
                        keywords = asset.keywords

                    # if a linked object is the target field, collect all keywords and search
                    # in case they already exist, link existing keywords and create only new ones
                    for k in keywords:
                        if k not in c.keyword_field_map:
                            continue

                        keyword_field = c.keyword_field_map[k]
                        if not k in linked_object_ids:
                            linked_object_ids[k] = {}

                        if keyword_field.is_linked():
                            if not k in obj['__keywords']:
                                obj['__keywords'][k] = []
                            obj['__keywords'][k] = []
                            for keyword_string in keywords[k]:
                                if not keyword_string in linked_object_ids[k]:
                                    linked_object_ids[k][keyword_string] = 0
                                obj['__keywords'][k].append(keyword_string)
                        else:
                            # determine if the target field is inside a nested table
                            if keyword_field.is_in_nested():
                                keyword_list = []

                                for keyword_string in keywords[k]:
                                    if keyword_field.is_l10n():
                                        keyword_list.append({
                                            keyword_field.name: {
                                                c.language: keyword_string
                                            }
                                        })
                                    else:
                                        keyword_list.append({
                                            keyword_field.name: keyword_string
                                        })

                                obj[c.objecttype][keyword_field.nested] = keyword_list
                            else:
                                keyword_string = ', '.join(keywords[k])
                                if keyword_field.is_l10n():
                                    obj[c.objecttype][keyword_field.name] = {
                                        c.language: keyword_string
                                    }
                                else:
                                    obj[c.objecttype][keyword_field.name] = keyword_string

                    if subject is not None and c.subject_field is not None:
                        if c.subject_field.is_l10n():
                            obj[c.objecttype][c.subject_field.name] = {
                                c.language: subject
                            }
                        else:
                            obj[c.objecttype][c.subject_field.name] = subject

                    for k in linked_object_ids:
                        # search all linked keyword objects where the collected keywords are set

                        if k not in c.keyword_field_map:
                            continue

                        keyword_field = c.keyword_field_map[k]

                        has_more_linked = True
                        limit = 1000
                        offset = 0

                        keywords_to_search = []
                        for kw in linked_object_ids[k]:
                            if linked_object_ids[k][kw] < 1:
                                keywords_to_search.append(kw)

                        if len(keywords_to_search) > 0:
                            while has_more_linked:
                                query = build_linked_object_search_query(
                                    keyword_field.linked_ot,
                                    keyword_field.name,
                                    list(keywords_to_search),
                                    language=c.language if keyword_field.is_l10n() else None,
                                    offset=offset,
                                    limit=limit)
                                result = do_search(self.easydb_context,
                                                   self.easydb_server,
                                                   self.easydb_token,
                                                   query)

                                hits = get_json_value(result, 'objects')
                                if hits is not None:
                                    size_linked_search_results = len(hits)

                                    self.logger.debug(
                                        '[%s] result: %d linked objects' % (c.debug_name, size_linked_search_results))
                                    if hits is not None:
                                        for hit in hits:
                                            ids = get_json_value(
                                                hit, '_fields.id')
                                            keywords = get_json_value(
                                                hit, '_fields.keyword')
                                            if ids is None or keywords is None:
                                                continue
                                            if len(ids) != len(keywords) or len(ids) != 1:
                                                continue
                                            try:
                                                linked_object_ids[k][keywords[0]] = int(
                                                    ids[0])
                                            except:
                                                continue
                                else:
                                    size_linked_search_results = 0

                                if size_linked_search_results < limit:
                                    has_more_linked = False
                                    break

                                offset += limit

                        # new linked objects for all keywords that have not been found
                        linked_objects_to_create = []
                        linked_event_infos = []
                        for kw in keywords_to_search:
                            if kw in linked_object_ids[k] and linked_object_ids[k][kw] > 0:
                                continue

                            linked_objects_to_create.append({
                                '_comment': 'inserted by auto-keyworder plugin with information from Cloudsight.ai',
                                '_mask': keyword_field.linked_ot_mask,
                                '_objecttype': keyword_field.linked_ot,
                                keyword_field.linked_ot: {
                                    '_version': 1,
                                    keyword_field.name: kw if not keyword_field.is_l10n() else {
                                        c.language: kw
                                    }
                                }
                            })

                        if len(linked_objects_to_create) > 0:
                            self.api_logger.info('[%s] insert %d linked objects of type %s' % (
                                c.debug_name,
                                len(linked_objects_to_create),
                                keyword_field.linked_ot
                            ))
                            self.api_logger.debug('[%s] post objects: %s' % (
                                c.debug_name, dumpjs(linked_objects_to_create)))

                            statuscode, resp = easydb_post_objects(self.easydb_server,
                                                                   self.easydb_token,
                                                                   keyword_field.linked_ot,
                                                                   linked_objects_to_create,
                                                                   format='long')
                            if statuscode == 200:
                                self.api_logger.debug('[%s] successfully created %d objects' % (
                                    c.debug_name,
                                    len(linked_objects_to_create)
                                ))

                                for ob in resp:
                                    _sys_id = get_json_value(
                                        ob, '_system_object_id')
                                    if _sys_id is None:
                                        continue
                                    _id = get_json_value(
                                        ob, '%s._id' % (keyword_field.linked_ot))
                                    if _id is None:
                                        continue
                                    kw = get_json_value(ob, '%s.%s%s' % (keyword_field.linked_ot,
                                                                         keyword_field.name,
                                                                         ('.%s' % c.language) if keyword_field.is_l10n() else ''))
                                    if kw is None:
                                        continue

                                    linked_object_ids[k][kw] = _id

                                    linked_event_infos.append({
                                        'object': {
                                            '_id': _id,
                                            '_system_object_id': _sys_id,
                                            '_objecttype': keyword_field.linked_ot,
                                            '_mask': keyword_field.linked_ot_mask
                                        },
                                        'language': c.language,
                                        'cloudsight_keyword': kw
                                    })

                                log_object_events(self.easydb_context,
                                                  linked_event_infos,
                                                  AUTO_KEYWORDER_OBJECT_CREATE)

                            else:
                                self.api_logger.error('%d: %s' %
                                                      (statuscode, dumpjs(resp)))

                    self.asset_status_map[sys_id].ai_data_added = True

                # finalize updated objects and generate event info
                for sys_id in self.asset_status_map:

                    asset = self.asset_status_map[sys_id]
                    if not asset.ai_data_added:
                        continue
                    if asset.updated:
                        continue
                    if asset.failed:
                        continue
                    if asset.subject is None and asset.keywords is None:
                        continue

                    obj = asset.easydb_obj
                    obj_id = get_json_value(obj, '%s._id' % c.objecttype, True)

                    _version = get_json_value(
                        obj, '%s._version' % c.objecttype)
                    if not isinstance(_version, int):
                        raise Exception(
                            'could not increase _version for object of type %s' % c.objecttype)
                    if _version < 1:
                        raise Exception('could not increase _version for object of type %s: invalid version %d' % (
                            c.objecttype, _version))

                    _version += 1
                    obj[c.objecttype]['_version'] = _version

                    self.logger.debug('[%s] update %s object %d (eas_id: %s) to version %d with subject: %s, %d keywords: [%s]' % (
                        c.debug_name,
                        c.objecttype,
                        obj_id,
                        asset.eas_id,
                        _version,
                        '"%s"' % asset.subject if asset.subject is not None else 'N/A',
                        len(keywords),
                        ', '.join(keywords)))

                    obj[c.objecttype][c.timestamp_field.name] = {
                        'value': format_datetime(dt.utcnow())
                    }

                    obj['_comment'] = 'Updated by auto-keyworder plugin with information from Cloudsight.ai'

                    event_info = {
                        'asset': {
                            'eas_id': asset.eas_id,
                            'extension': asset.img_type,
                            'original_filepath': asset.original_filepath,
                            'version': c.asset_version
                        },
                        'object': {
                            '_id': obj_id,
                            '_system_object_id': sys_id,
                            '_objecttype': c.objecttype
                        },
                        'language': c.language
                    }

                    if asset.subject is not None:
                        event_info['cloudsight_subject'] = asset.subject

                    # iterate over keywords, link to searched / newly created keyword objects
                    for k in linked_object_ids:

                        if k not in c.keyword_field_map:
                            continue

                        linked_keywords = get_json_value(
                            obj, '__keywords.%s' % k)
                        if not isinstance(linked_keywords, list) or len(linked_keywords) < 1:
                            continue

                        keyword_field = c.keyword_field_map[k]

                        # create an empty nested table for the linked objects
                        nested = [
                            {
                                '__keyword': kw,
                                keyword_field.link_name: {
                                    keyword_field.linked_ot: {
                                        '_id': linked_object_ids[k][kw]
                                    },
                                    '_objecttype': keyword_field.linked_ot,
                                    '_mask': keyword_field.linked_ot_mask
                                }
                            }
                            for kw in linked_keywords
                        ]
                        nested.sort(key=lambda x: x['__keyword'])
                        obj[c.objecttype][keyword_field.nested] = nested

                        if not 'cloudsight_keywords' in event_info:
                            event_info['cloudsight_keywords'] = {}

                        event_info['cloudsight_keywords'][k] = []
                        for kw in linked_keywords:
                            event_info['cloudsight_keywords'][k].append({
                                'keyword': kw,
                                '_objecttype': keyword_field.linked_ot,
                                '_id': linked_object_ids[k][kw]
                            })

                    objects_to_update.append(obj)
                    update_event_infos.append(event_info)

                    self.asset_status_map[sys_id].updated = True

                # post the list of objects to api/v1/db/<objecttype>
                # if update was successful, log events for each object
                if len(objects_to_update) > 0:
                    self.api_logger.info('[%s] update %d objects of type %s' % (
                        c.debug_name,
                        len(objects_to_update),
                        c.objecttype
                    ))
                    self.api_logger.debug('[%s] post objects: %s' % (
                        c.debug_name, dumpjs(objects_to_update)))
                    statuscode, resp = easydb_post_objects(self.easydb_server,
                                                           self.easydb_token,
                                                           c.objecttype,
                                                           objects_to_update)
                    if statuscode == 200:
                        self.api_logger.debug('[%s] successfully updated %d objects' % (
                            c.debug_name,
                            len(objects_to_update)
                        ))
                        log_object_events(self.easydb_context,
                                          update_event_infos,
                                          AUTO_KEYWORDER_OBJECT_UPDATE)
                    else:
                        self.api_logger.error('%d: %s' %
                                              (statuscode, dumpjs(resp)))

                if size_search_results < self.search_chunk_size:
                    self.logger.debug('[%s] size %d < %d => break search loop' % (
                        c.debug_name,
                        size_search_results,
                        self.search_chunk_size))
                    has_more = False

                offset += self.search_chunk_size


class CloudsightKeywordCollector(object):

    def __init__(self, logger, easydb_context, api_key, language, eas_id, objecttype, obj_sid, base64_image, img_type, original_filepath, uploaded, keyword_map_keys):
        self.logger = logger
        self.api_key = api_key
        self.language = language
        self.eas_id = eas_id
        self.objecttype = objecttype
        self.obj_sid = obj_sid
        self.base64_image = base64_image
        self.img_type = img_type
        self.original_filepath = original_filepath
        self.uploaded = uploaded
        self.keyword_map_keys = keyword_map_keys
        self.easydb_context = easydb_context
        self.CLOUDSIGHT_URL = 'https://api.cloudsight.ai/v1/images'
        self.cloudsight_token = None

    def build_cloudsight_auth_headers(self):
        return {
            'authority': 'api.cloudsight.ai',
            'authorization': 'CloudSight ' + self.api_key
        }

    def parse_cloudsight_response(self, resp_js):

        structured_output_keys = {
            'quantity': {
                'de': 'menge'
            },
            'gender': {
                'de': 'geschlecht'
            },
            'material': {
                'de': 'material'
            },
            'color': {
                'de': 'farbe'
            }
        }

        self.logger.debug('cloudsight response for asset %d in object #%d: %s' %
                          (self.eas_id, self.obj_sid, dumpjs(resp_js)))

        # cloudsight wont return any keywords for this image
        skipped = get_json_value(resp_js, 'skipped')
        if skipped is not None:
            self.logger.debug('cloudsight skipped asset %d in object #%d, reason: %s' %
                              (self.eas_id, self.obj_sid, str(skipped)))
            raise AssetSkippedException(skipped)

        # get status, check if keywords can be returned or the request must be repeated later
        status = get_json_value(resp_js, 'status')
        if status == 'completed':

            subject = None
            name = get_json_value(resp_js, 'name')
            if is_string_set(name):
                subject = name

            keywords = {}

            arr_name = 'categories'
            map_key = 'keyword_categories_field'
            if map_key in self.keyword_map_keys:
                keywords[map_key] = []
                arr = get_json_value(resp_js, arr_name)
                if isinstance(arr, list):
                    for keyword in arr:
                        if is_string_set(keyword) and keyword not in keywords[map_key]:
                            keywords[map_key].append(keyword)
                if len(keywords[map_key]) < 1:
                    del keywords[map_key]
                else:
                    keywords[map_key].sort()

            structured = get_json_value(resp_js, 'structured_output')
            if isinstance(structured, dict):
                for a in structured_output_keys:
                    arr_name = a

                    # if not english, check if the key was translated
                    if self.language != 'en':
                        if self.language in structured_output_keys[a]:
                            arr_name = structured_output_keys[a][self.language]

                    arr = get_json_value(structured, arr_name)
                    if not isinstance(arr, list) or len(arr) < 1:
                        continue

                    self.logger.debug(
                        'values for structured.%s: %s' % (arr_name, str(arr)))

                    map_key = None
                    if a == 'quantity' and 'keyword_quantity_field' in self.keyword_map_keys:
                        map_key = 'keyword_quantity_field'
                    elif a == 'gender' and 'keyword_gender_field' in self.keyword_map_keys:
                        map_key = 'keyword_gender_field'
                    elif a == 'material' and 'keyword_material_field' in self.keyword_map_keys:
                        map_key = 'keyword_material_field'
                    elif a == 'color' and 'keyword_color_field' in self.keyword_map_keys:
                        map_key = 'keyword_color_field'
                    if map_key is None:
                        continue

                    keywords[map_key] = []
                    for keyword in arr:
                        if is_string_set(keyword) and keyword not in keywords[map_key]:
                            keywords[map_key].append(keyword)
                    keywords[map_key].sort()

            # subject, keywords, finished, failed
            return subject, keywords, True, False

        # not completed yet, return only token
        # subject => token, keywords, finished, failed
        return get_json_value(resp_js, 'token'), None, False, False

    def upload_image_to_cloudsight(self, base64_image, img_type, obj_id, asset_version):
        _headers = self.build_cloudsight_auth_headers()
        self.logger.debug('cloudsight request for asset %d in object #%d: POST %s' %
                          (self.eas_id, self.obj_sid, dumpjs({
                              'image': 'data:image/' + img_type + ';base64,...',
                              'language': self.language
                          })))
        response = requests.post(
            self.CLOUDSIGHT_URL,
            headers=_headers,
            data={
                'image': 'data:image/' + img_type + ';base64,' + base64_image,
                'language': self.language
            })
        log_event(self.easydb_context,
                  AUTO_KEYWORDER_CLOUDSIGHT_REQUEST,
                  {
                      'asset': {
                          'eas_id': self.eas_id,
                          'extension': self.img_type,
                          'original_filepath': self.original_filepath,
                          'version': asset_version
                      },
                      'object': {
                          '_id': obj_id,
                          '_objecttype': self.objecttype,
                          '_system_object_id': self.obj_sid
                      },
                      'request': {
                          'method': 'POST',
                          'url': self.CLOUDSIGHT_URL,
                          'body': {
                              'language': self.language
                          }
                      },
                      'response': {
                          'statuscode': response.status_code,
                          'body': response.content.decode('utf-8')
                      }
                  })

        if response.status_code != 200:
            # subject, keywords, finished, failed
            return None, None, False, True

        return self.parse_cloudsight_response(json.loads(response.content.decode('utf-8')))

    def get_keywords_from_cloudsight(self, obj_id, asset_version):
        if self.cloudsight_token is None:
            self.logger.debug('could not get keywords for asset %d in object #%d: no cloudsight token' % (
                self.eas_id, self.obj_sid))
            # subject, keywords, finished, failed
            return None, None, False, True

        _headers = self.build_cloudsight_auth_headers()
        self.logger.debug('cloudsight request for asset %d in object #%d: GET %s' %
                          (self.eas_id, self.obj_sid, dumpjs({
                              'token': self.cloudsight_token
                          })))
        response = requests.get(
            self.CLOUDSIGHT_URL + '/' + self.cloudsight_token,
            headers=_headers)
        log_event(self.easydb_context,
                  AUTO_KEYWORDER_CLOUDSIGHT_REQUEST,
                  {
                      'asset': {
                          'eas_id': self.eas_id,
                          'extension': self.img_type,
                          'original_filepath': self.original_filepath,
                          'version': asset_version
                      },
                      'object': {
                          '_id': obj_id,
                          '_objecttype': self.objecttype,
                          '_system_object_id': self.obj_sid
                      },
                      'request': {
                          'method': 'GET',
                          'url': self.CLOUDSIGHT_URL + '/' + self.cloudsight_token
                      },
                      'response': {
                          'statuscode': response.status_code,
                          'body': response.content.decode('utf-8')
                      }
                  })

        if response.status_code != 200:
            # subject, keywords, finished, failed
            return None, None, False, True

        return self.parse_cloudsight_response(json.loads(response.content.decode('utf-8')))

    def request_keywords(self, obj_id, asset_version):

        subject = None
        keywords = None
        finished = False
        failed = False

        try:
            if not self.uploaded:
                self.uploaded = True

                if self.img_type is None:
                    self.img_type = 'jpg'
                    self.logger.debug(
                        'no image type from asset info, use fallback %s' % (self.img_type))

                self.logger.debug('upload image, eas id: %d, type: %s' %
                                  (self.eas_id, self.img_type))
                subject, keywords, finished, failed = self.upload_image_to_cloudsight(
                    self.base64_image, self.img_type, obj_id, asset_version)

                if not finished and not failed:
                    self.cloudsight_token = subject
                    subject = None
            else:
                self.logger.debug('get status of image, eas id: %d, token: %s' % (
                    self.eas_id, self.cloudsight_token))
                subject, keywords, finished, failed = self.get_keywords_from_cloudsight(
                    obj_id, asset_version)

            # subject, keywords, finished, failed, reason
            return subject, keywords, finished, failed, None

        except AssetSkippedException as e:
            event_info = self.format_event_info(obj_id, asset_version)
            event_info['error'] = 'asset was skipped by cloudsight'
            if asset.reason is not None:
                event_info['details'] = 'reason: %s' % (asset.reason)
            log_event(self.easydb_context,
                      AUTO_KEYWORDER_ERROR,
                      event_info)

            # subject, keywords, finished, failed, reason
            return None, None, True, True, e.reason

        except requests.exceptions.ConnectionError as e:
            event_info = self.format_event_info(obj_id, asset_version)
            event_info['error'] = 'could not connect to cloudsight server %s' % (
                self.CLOUDSIGHT_URL)
            event_info['details'] = str(e.message)
            log_event(self.easydb_context,
                      AUTO_KEYWORDER_ERROR,
                      event_info)

            # subject, keywords, finished, failed, reason
            return None, None, True, True, None

        except Exception as e:
            trace = print_traceback(self.logger, e)

            event_info = self.format_event_info(obj_id, asset_version)
            event_info['error'] = 'generic'
            event_info['details'] = trace
            log_event(self.easydb_context,
                      AUTO_KEYWORDER_ERROR,
                      event_info)

            # subject, keywords, finished, failed, reason
            return None, None, True, True, None

    def format_event_info(self, obj_id, asset_version):
        return {
            'asset': {
                'eas_id': self.eas_id,
                'extension': self.img_type,
                'original_filepath': self.original_filepath,
                'version': asset_version
            },
            'object': {
                '_objecttype': self.objecttype,
                '_id': obj_id,
                '_system_object_id': self.obj_sid
            }
        }
